# fmt: off
# autopep8: off
from ._version import get_versions

__version__ = get_versions()['version']
del get_versions



import os

import intake

from .datasources import MetOfficeDataSource, MergedMetOfficeDataSource

CATALOG_DIR = os.path.join(os.path.dirname(__file__), 'cats')

mogreps_cat = intake.open_catalog(os.path.join(CATALOG_DIR, 'mogreps_uk_cat.yaml'))
aq_cat = intake.open_catalog(os.path.join(CATALOG_DIR, 'air_quality_cat.yaml'))
ukv_2020_plus_cat = intake.open_catalog(os.path.join(CATALOG_DIR, 'ukv_timeseries.yaml'))

cat = intake.catalog.Catalog.from_dict(
    {
        'air_quality':aq_cat,
        'weather_forecasts':mogreps_cat,
        'weather_timeseries_uk':ukv_2020_plus_cat

    }, name="Met Office Datasets")

# fmt: on
# autopep8: on



def timedelta_to_duration_str(td):
    """Convert a timedelta object into a duration string e.g: PT<hhhh>H<mm>M."""
    minutes, seconds = divmod(td.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    hours += td.days * 24
    hours = int(hours)
    assert seconds == 0
    return f"PT{hours:04}H{minutes:02}M"


def datetime_to_iso_str(dt):
    """Convert a datetime object into an ISO string representation."""
    return dt.strftime("%Y%m%dT%H%MZ")


def calc_cycle_validity_lead_times(cycle_time=None, validity_time=None, lead_time=None):
    """Given two of cycle time, validity time and lead time, return all three."""
    num_inputs = sum(
        map(lambda x: 1 if x is not None else 0, [cycle_time, validity_time, lead_time])
    )
    if num_inputs != 2:
        raise RuntimeError("Expected 2 of cycle_time, validity_time and lead_time")
    if not validity_time:
        validity_time = cycle_time + lead_time
    elif not cycle_time:
        cycle_time = validity_time - lead_time
    elif not lead_time:
        lead_time = validity_time - cycle_time
    return cycle_time, validity_time, lead_time



import datetime

from intake.catalog.local import YAMLFilesCatalog
from intake.source.base import Schema
from intake_xarray.base import DataSourceMixin
from met_office_datasets import __version__

from .dataset import MODataset
from .utils import datetime_to_iso_str

DATA_DELAY = 24  # num hours from current time that data is available


class MetOfficeDataSource(DataSourceMixin):
    name = "met_office"
    version = __version__

    def __init__(
        self,
        start_cycle,
        end_cycle,
        cycle_frequency,
        forecast_extent,
        model,
        dimensions,
        diagnostics,
        static_coords,
        storage_options,
        metadata=None,
    ):
        super().__init__(metadata=metadata)
        if end_cycle.lower() == "latest":
            end_cycle = datetime.datetime.utcnow() - datetime.timedelta(
                hours=DATA_DELAY
            )
            end_cycle = datetime_to_iso_str(end_cycle)

        self.start_cycle = start_cycle
        self.end_cycle = end_cycle
        self.cycle_frequency = cycle_frequency
        self.forecast_extent = forecast_extent
        self.model = model
        self.dimensions = dimensions
        self.diagnostics = diagnostics
        self.static_coords = static_coords
        self.storage_options = storage_options
        self._ds = None

    def _open_dataset(self):
        self._ds = MODataset(
            start_cycle=self.start_cycle,
            end_cycle=self.end_cycle,
            model=self.model,
            dims=self.dimensions,
            diagnostics=self.diagnostics,
            static_coords=self.static_coords,
            cycle_freq=self.cycle_frequency,
            start_lead_time="0H",
            end_lead_time=self.forecast_extent,
            lead_time_freq="1H",
            **self.storage_options
        ).ds

    def _get_schema(self):
        # adapted from intake-xarray driver
        if self._ds is None:
            self._open_dataset()

            # assume rectangular data (shared coords across all data vars)
            metadata = {
                "dims": dict(self._ds.dims),
                "data_vars": self.diagnostics,
                "coords": tuple(self._ds.coords.keys()),
            }
            metadata.update(self._ds.attrs)
            self._schema = Schema(
                datashape=None,
                dtype=None,
                shape=None,
                npartitions=None,
                extra_metadata=metadata,
            )
        return self._schema


class MergedMetOfficeDataSource(YAMLFilesCatalog):
    name = "merged_met_office"
    version = __version__

    def __init__(self, path, flatten=True, **kwargs):
        metadata = kwargs.pop("metadata")
        super().__init__(path, flatten=flatten, metadata=metadata)
        self._kwargs = kwargs
        self._ds = None

    def to_dask(self):
        if self._ds is not None:
            return self._ds

        for entry in self._entries.values():
            dataset = entry(**self._kwargs).to_dask()
            if self._ds is None:
                self._ds = dataset
            else:
                self._ds = self._ds.merge(dataset)
        return self._ds

    def read_chunked(self):
        return self.to_dask()



import json
from collections import namedtuple
from collections.abc import MutableMapping
from functools import partial
from itertools import product

import numpy as np
import xarray as xr
import zarr

FlagsProxy = namedtuple("FlagsProxy", ("c_contiguous",))
FlagsProxy.__new__.__defaults__ = (True,)

# values would ordinarily be an np array, so would contain dtype,
# shape etc - we dummy this just to have access to the is_c_contiguous flag
ValuesProxy = namedtuple("ValuesProxy", ("flags",))
ValuesProxy.__new__.__defaults__ = (FlagsProxy(),)

# data proxy to access chunksize
DataProxy = namedtuple("DataProxy", ("chunksize",))
DataProxy.__new__.__defaults__ = (None,)

# constructor signature not the same, but gives access to the attributes
# that we need
VariableProxy = namedtuple(
    "VariableProxy", ("dims", "shape", "dtype", "values", "data", "attrs")
)
VariableProxy.__new__.__defaults__ = (ValuesProxy(), DataProxy(), {})


class HypotheticZarrStore(MutableMapping):
    # For now define only the MutableMapping methods which are
    # overridden for zarr.storage.DirectoryStore
    # Could also implement:
    #     # mixins from base class
    #     def __contains__(self, item):
    #     def items(self):
    #     def values(self):
    #     def get(self, key, default=None):
    #     def __ne__(self, other):
    #     # added through mixin
    #     def pop(self, key, default=None):
    #     def popitem(self, kv):
    #     def update(self, **kwargs):
    #     def setdefault(self, key, default=None):
    METADATA_KEY = ".zmetadata"
    META_KEYS = [
        zarr.storage.array_meta_key,
        zarr.storage.attrs_key,
    ]
    ROOT_KEYS = [
        zarr.storage.group_meta_key,
        zarr.storage.attrs_key,
        METADATA_KEY,
    ]

    def __init__(
        self,
        dims,
        coord_vars,
        data_vars,
        chunks,
        loader_function,
        attrs=None,
        dtypes=None,
    ):
        # dims is a list/tuple of strs
        # coord vars is a dictionary of variables
        # datavars is a list of strs (assume all have the same dims -> same shape)
        # chunks is a dict describing chunks in a file (if missing, assume value is 1)
        # loader_function goes from coord values to data (via file)
        #   func assumed to return np array

        # optional:
        # attrs is a dict of global attrs for whole dataset
        # dtypes is a dict containing strings which specify the numpy dtype
        # of the data in the array - if not specified, float32 assumed

        # guard clause
        assert all(map(lambda dim: dim in coord_vars, dims))
        self.dims = dims
        self.attrs = attrs if attrs else {}
        self.chunks = chunks
        self.loader_function = loader_function

        # converting coord (data) arrays to dask arrays
        # to have access to number of chunks (put whole array in one chunk)
        self.coord_vars = {
            name: coord.copy().chunk() for name, coord in coord_vars.items()
        }
        if dtypes is None:
            dtypes = {}
        # need to do this after dims, chunks, coord_vars defined
        self.data_vars = {
            name: self._create_var_proxy(dtypes.get(name, "float32"))
            for name in data_vars
        }

    @property
    def vars(self):
        return dict(self.coord_vars, **self.data_vars)

    def _create_var_proxy(self, dtype, attrs=None, c_contiguous=True):
        if attrs is None:
            attrs = {}
        # dtype is a str passed into np.dtype()
        # calc: all data_vars have the same dims => same shape
        shape = tuple(map(lambda dim: len(self.coord_vars[dim]), self.dims))

        # get from dict, assume chunk of 1 if not present
        chunksize = self._chunksize(self.chunks, self.dims)
        # fake values to give acess to c_contiguous flag
        data = DataProxy(chunksize)
        values = ValuesProxy(FlagsProxy(c_contiguous))
        return VariableProxy(
            dims=self.dims,
            shape=shape,
            dtype=np.dtype(dtype),
            data=data,
            values=values,
            attrs=attrs,
        )

    @staticmethod
    def _chunksize(chunks, dims):
        return tuple(chunks.get(dim, 1) for dim in dims)

    @staticmethod
    def _num_chunks(var):
        return tuple(s // c for s, c in zip(var.shape, var.data.chunksize))

    @staticmethod
    def _var_mem_order(var):
        return "C" if var.values.flags.c_contiguous else "F"

    def _zgroup_dict(self):
        return json.dumps({"zarr_format": 2})

    def _zattrs_dict(self, variable=None):
        if variable is None:
            return json.dumps(self.attrs)
        else:
            return json.dumps(
                dict({"_ARRAY_DIMENSIONS": variable.dims}, **variable.attrs)
            )

    def _zarray_dict(self, variable):
        return json.dumps(
            {
                "chunks": variable.data.chunksize,
                "compressor": None,
                "dtype": variable.dtype.str,
                "fill_value": None,
                "filters": None,
                "order": self._var_mem_order(variable),
                "shape": variable.shape,
                "zarr_format": 2,
            }
        )

    def _zmetadata_dict(self):
        # if called from __getitem__, recursive (depth=2) and converts back and forth
        # between strings and dicts using json lib - should not be too slow though
        metadata = {}
        for key in filter(lambda key: key != self.METADATA_KEY, self.ROOT_KEYS):
            metadata[key] = json.loads(self[key].decode())
        for var in self.vars:
            for key in self.META_KEYS:
                key = f"{var}/{key}"
                # convert back from bytestring to regular string
                metadata[key] = json.loads(self[key].decode())
        return json.dumps({"metadata": metadata, "zarr_consolidated_format": 1})

    def _get_dim_values(self, chunk_idxs, dims):
        chunksize = self._chunksize(self.chunks, dims)
        # move to data index space rather than chunk indexing..
        data_idxs = list(map(lambda x: x[0] * x[1], zip(chunk_idxs, chunksize)))
        values = []
        for dim, idx in zip(dims, data_idxs):
            val = self.coord_vars[dim].values[idx]
            values.append(val.item())
        mapping = dict(zip(dims, values))
        return mapping

    def __getitem__(self, item):
        key = item.split("/")
        if len(key) == 1:
            var_name = None
            key = key[0]
        elif len(key) == 2:
            var_name, key = key
        else:
            raise KeyError(f"Invalid key: {item}")
        # mapping from keys to metadata returning methods
        accessor_mapping = {
            zarr.storage.array_meta_key: self._zarray_dict,
            zarr.storage.attrs_key: self._zattrs_dict,
            zarr.storage.group_meta_key: self._zgroup_dict,
            self.METADATA_KEY: self._zmetadata_dict,
        }
        # getting metadata:
        if key in accessor_mapping:
            accessor = accessor_mapping[key]
            if var_name is not None:
                accessor = partial(accessor, self.vars[var_name])
            return accessor().encode()
        # getting coords (stored in object):
        var = self.vars[var_name]
        if var_name in self.coord_vars:
            data = var.values
        else:
            # getting data (from elsewhere):
            accessor = self.loader_function
            # returning data - key are chunk indices
            chunk_idxs = tuple(int(x) for x in key.split("."))
            # access_values are the starting values for that chunk in all dims
            # as well as the variable name (first part of the key)
            access_values = self._get_dim_values(chunk_idxs, var.dims)
            access_values["variable_name"] = var_name
            data = accessor(access_values)
            if not isinstance(data, np.ndarray) and data is not None:
                raise TypeError("Loader function should return a numpy.ndarray or None")
            if data is None:
                data = np.full(
                    shape=var.data.chunksize,
                    fill_value=np.nan,
                    dtype=var.dtype,
                    order=self._var_mem_order(var),
                )
            # could potentially do some checking that shape and dtype are as expected if loaded
        return data.tobytes(order=self._var_mem_order(var))

    def __setitem__(self, item, value):
        raise NotImplementedError("Read-only access provided.")

    def __delitem__(self, item):
        raise NotImplementedError("Read-only access provided.")

    def __iter__(self):
        return self.keys()

    def __len__(self):
        # slow, could be calculated...
        # for now use zarr.storage.DirectoryStore implementation
        return sum(1 for _ in self.keys())

    def keys(self):
        for key in self.ROOT_KEYS:
            yield key

        for name, var in self.vars.items():
            for key in self.META_KEYS:
                yield f"{name}/{key}"
            num_chunks = self._num_chunks(var)
            chunk_iters = [range(x) for x in num_chunks]
            # slow, likely not to be used often
            # could calculate faster (at the cost of more memory)
            # by using np.meshgrid for cartesian products etc.
            for chunk_idx in product(*chunk_iters):
                chunk_idx = ".".join(str(x) for x in chunk_idx)
                yield f"{name}/{chunk_idx}"

    @staticmethod
    def _xarray_variables_equal(var1, var2):
        return (
            var1.dims == var2.dims
            and var1.attrs == var2.attrs
            and np.array_equal(var1.values, var2.values)
        )

    def __eq__(self, other):
        # only do some basic checking here.. could do more
        # could speed up by exiting early if any evaluates to False
        correct_type = isinstance(other, HypotheticZarrStore)
        dims_same = self.dims == other.dims
        data_vars_same = self.data_vars == other.data_vars
        coord_var_keys_same = self.coord_vars.keys() == other.coord_vars.keys()
        coord_var_values_same = all(
            [
                self._xarray_variables_equal(self_var, other_var)
                for self_var, other_var in zip(
                    self.coord_vars.values(), other.coord_vars.values()
                )
            ]
        )
        coord_vars_same = coord_var_keys_same and coord_var_values_same
        return correct_type and dims_same and data_vars_same and coord_vars_same

    def clear(self):
        raise NotImplementedError("Read-only access provided.")


class HypotheticZarrCloner(HypotheticZarrStore):
    def __init__(self, target, *args, **kwargs):
        if not isinstance(target, MutableMapping):
            raise ValueError("Target must be a MutableMapping")
        self.target = target
        super().__init__(*args, **kwargs)

    # keeps chunking exactly the same
    def __getitem__(self, item):
        try:
            return self.target[item]
        except KeyError:
            value = super().__getitem__(item)
            self.target[item] = value
            return value



from .dataset import MODataset
import xarray as xr
import numpy as np
import pandas as pd
from .utils import datetime_to_iso_str
from io import BytesIO
import datetime
from met_office_datasets import __version__
from met_office_datasets.datasources import MetOfficeDataSource


class TimeSeriesDatasource(MetOfficeDataSource):
    name = "met_office_ukv_timeseries"
    version = __version__

    def __init__(self,
                 start_cycle,
                 end_cycle,
                 cycle_frequency,
                 model,
                 dimensions,
                 diagnostics,
                 static_coords,
                 storage_options,
                 metadata=None
                 ):

        if end_cycle.lower() == 'latest':
            end_cycle = datetime_to_iso_str((datetime.datetime.now() - datetime.timedelta(hours=48)))

        super().__init__(
            start_cycle=start_cycle,
            end_cycle=end_cycle,
            cycle_frequency=cycle_frequency,
            forecast_extent=None,
            model=model,
            dimensions=dimensions,
            static_coords=static_coords,
            diagnostics=diagnostics,
            storage_options=storage_options,
            metadata=metadata
        )

    def _open_dataset(self):
        self._ds = TimeSeriesDataset(
            start_cycle=self.start_cycle,
            end_cycle=self.end_cycle,
            model=self.model,
            dims=self.dimensions,
            diagnostics=self.diagnostics,
            static_coords=self.static_coords,
            cycle_frequency=self.cycle_frequency,
            storage_options=self.storage_options,
        ).ds


class MetOfficeAQDataSource(MetOfficeDataSource):
    name = "met_office_aq"
    version = __version__

    def __init__(self,
                 start_cycle,
                 end_cycle,
                 cycle_frequency,
                 model,
                 dimensions,
                 diagnostics,
                 static_coords,
                 storage_options,
                 aggregation=None,
                 metadata=None
                 ):

        if end_cycle.lower() == 'latest':
            end_cycle = datetime_to_iso_str((datetime.datetime.now() - datetime.timedelta(hours=48)))

        self.aggregation = aggregation

        super().__init__(
            start_cycle=start_cycle,
            end_cycle=end_cycle,
            cycle_frequency=cycle_frequency,
            forecast_extent=None,
            model=model,
            dimensions=dimensions,
            static_coords=static_coords,
            diagnostics=diagnostics,
            storage_options=storage_options,
            metadata=metadata
        )

    def _open_dataset(self):
        self._ds = AQDataset(
            start_cycle=self.start_cycle,
            end_cycle=self.end_cycle,
            model=self.model,
            dims=self.dimensions,
            diagnostics=self.diagnostics,
            static_coords=self.static_coords,
            cycle_frequency=self.cycle_frequency,
            storage_options=self.storage_options,
            aggregation=self.aggregation
        ).ds


class SingleTimeDataset(MODataset):

    def __init__(
        self,
        start_cycle,
        end_cycle,
        model,
        dims,
        diagnostics,
        static_coords,
        cycle_frequency,
        storage_options,
        aggregation=None
    ):

        super().__init__(
            start_cycle=start_cycle,
            end_cycle=end_cycle,
            model=model,
            dims=dims,
            diagnostics=diagnostics,
            static_coords=static_coords,
            cycle_freq=cycle_frequency,
            start_lead_time=None,
            end_lead_time=None,
            lead_time_freq=None,
            **storage_options)

        self.aggregation = aggregation

    @ property
    def chunks(self):
        static_coords = self.static_coords
        chunks = {name: static_coords[name].shape[0] for name in static_coords.keys()}
        assert self.cycle_freq in ['1H', '1D']
        if self.cycle_freq == '1H':
            time_chunks = 24
        elif self.cycle_freq == '1D':
            time_chunks = 1
        else:
            raise RuntimeError(f"Don't know how to deal with cycle_freq {cycle_freq} for chunking")

        chunks.update({'time': time_chunks})
        return chunks

    @staticmethod
    def _check_dims_coords(dims, static_coords, model):
        # two types of grid def, assume one based on presence of "grid_latitude" or not
        if "grid_latitude" in dims:
            expected_coords = [
                "grid_longitude",
                "grid_latitude"
            ]
        else:
            expected_coords = [
                "projection_x_coordinate",
                "projection_y_coordinate"
            ]

        expected_dims = [
            "time"
        ] + expected_coords

        pair_dict = {
            "static_coords": (expected_coords, list(static_coords.keys())),
            "dims": (expected_dims, dims),
        }

        for var_type, pair in pair_dict.items():
            expected, passed_in = pair
            # check all of expected were passed in
            if not all(map(lambda var: var in passed_in, expected)):
                raise ValueError(f"Expected to find all of {expected} in {var_type}")

    @ property
    def dynamic_coords(self):
        dynamic_coords_data = {
            "time": pd.date_range(
                start=self.start_cycle, end=self.end_cycle, freq=self.cycle_freq
            )
        }
        return {
            name: xr.Variable(dims=(name,), data=data)
            for name, data in dynamic_coords_data.items()
        }

    def _zstore_loader(self, attrs):
        time = attrs["time"]
        diag = attrs["variable_name"]

        time = pd.to_datetime(np.datetime64(time, "ns"))

        url = self._get_blob_url(
            diagnostic=diag, time=time
        )

        try:
            data = self._read_from_url(url)
            data = xr.open_dataset(BytesIO(data))
            data = self._extract_data_as_dataarray(data)
            return data.values
        except FileNotFoundError:
            return None

    @ staticmethod
    def _extract_data_as_dataarray(dataset):
        # coords in all datasets
        DIM_COORD_VARS = [
            "time",
            "projection_y_coordinate",
            "projection_x_coordinate",
            "forecast_reference_time",
            "forecast_day",
            "grid_longitude",
            "grid_latitude",
            "forecast_period"
        ]

        # coords only in some datasets
        VERTICAL_COORD_VARS = ["height", "depth", "pressure"]

        # variables that are not in coords or dims (including 'bnds' in dims)
        # but in all datasets
        NON_COORD_DIM_VARS = [
            "lambert_azimuthal_equal_area",
            "projection_y_coordinate_bnds",
            "projection_x_coordinate_bnds",
            "transverse_mercator",
            "experiment_number",
            "rotated_latitude_longitude"
        ]

        # vars that are not coords/dims but are optional
        TIME_BOUND_VARS = ["time_bnds", "forecast_period_bnds", "forecast_reference_time_bnds"]

        # also has 'depth_bnds' optionally..
        # need to figure out which variable is the data variable
        NON_DATA_VARS = (
            DIM_COORD_VARS
            + VERTICAL_COORD_VARS
            + NON_COORD_DIM_VARS
            + TIME_BOUND_VARS
            + ["depth_bnds"]
        )

        data_var = list(set(dataset.variables) - set(NON_DATA_VARS))
        if len(data_var) > 1:
            raise RuntimeError("Expected to find only 1 data variable")
        data_var = data_var[0]
        return dataset[data_var]


class TimeSeriesDataset(SingleTimeDataset):

    def _get_blob_url(
        self, diagnostic, time=None
    ):
        """Return the URL of a forecast file."""

        # convert all to strings
        if self.cycle_freq == '1H':
            frequency = 'hourly'
        elif self.cycle_freq == '1D':
            frequency = 'daily'
        else:
            raise RuntimeError(f"Don't know how to deal with cycle_freq {cycle_freq} for chunking")

        time_str = datetime_to_iso_str(time).split('T')[0]
        obj_path = (
            f"metoffice_{self.model}_{frequency}/{diagnostic}/{self.model}_{frequency}_{diagnostic}_{time_str}.nc"
        )
        obj_path = f"{self.url_prefix}/{obj_path}"
        return f"{self.data_protocol}://{obj_path}"


class AQDataset(SingleTimeDataset):

    def _get_blob_url(
        self, diagnostic, time=None
    ):
        """Return the URL of a forecast file."""

        # convert all to strings
        time_str = datetime_to_iso_str(time).split('T')[0]
        ag_str = f"_{self.aggregation}" if self.aggregation else ""
        obj_path = (
            f"metoffice_{self.model}/{diagnostic}/{self.model}_{diagnostic}{ag_str}_{time_str}.nc"
        )
        obj_path = f"{self.url_prefix}/{obj_path}"
        return f"{self.data_protocol}://{obj_path}"



import versioneer
from setuptools import setup, find_packages
from os import path

this_dir = path.abspath(path.dirname(__file__))
with open(path.join(this_dir, "README.md")) as f:
    long_description = f.read()

NAME = "met_office_datasets"

setup(
    name=NAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="DESCRIPTION",
    url="https://github.com/informatics-lab/met_office_datasets",
    author="Theo McCaie",
    author_email="theo.mccaie@informaticslab.co.uk",
    classifiers=[
        # "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Natural Language :: English",
    ],
    packages=find_packages(),
    install_requires=["numpy>=1.11",
                      'dask>=1.0',
                      'xarray',
                      'zarr>=2',
                      'adlfs',
                      'h5netcdf>=0.8',
                      'adlfs',
                      'intake',
                      'intake-xarray',
                      'toolz'],
    zip_safe=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    entry_points={
        'intake.catalogs': [
            f'met_office= {NAME}:cat',
        ]
    }
)



from io import BytesIO

import fsspec
import numpy as np
import pandas as pd
import xarray as xr

from ..zarrhypothetic.zarrhypothetic import HypotheticZarrStore
from .utils import (
    calc_cycle_validity_lead_times,
    datetime_to_iso_str,
    timedelta_to_duration_str,
)


# TODO: remove hardcoded assumptions about MOGREPS-UK
class MODataset:
    def __init__(
        self,
        start_cycle,
        end_cycle,
        model,
        dims,
        diagnostics,
        static_coords,
        cycle_freq="1H",
        start_lead_time="0H",
        end_lead_time="126H",
        lead_time_freq="1H",
        **storage_options,
    ):
        """
        storage_options must contain the keys data_protocol and url_prefix

        NB: If using abfs, storage_options should contain the keys
        'account_name' and 'credential'
        """

        self._check_dims_coords(dims, static_coords, model)

        self.start_cycle = start_cycle
        self.end_cycle = end_cycle
        self.model = model
        self.dims = dims
        self.diagnostics = diagnostics
        self._static_coords = static_coords
        self.cycle_freq = cycle_freq
        self.start_lead_time = start_lead_time
        self.end_lead_time = end_lead_time
        self.lead_time_freq = lead_time_freq

        self.data_protocol = storage_options.pop("data_protocol")
        self.url_prefix = storage_options.pop("url_prefix")
        self.storage_options = storage_options
        self._validate_storage_options()

        # remove the 'Z' from the start/end points or xarray struggles...
        remove_trailing_z = (
            lambda dt_str: dt_str[:-1] if dt_str.endswith("Z") else dt_str
        )
        self.start_cycle = remove_trailing_z(self.start_cycle)
        self.end_cycle = remove_trailing_z(self.end_cycle)

        self._zstore = self._create_zstore()
        self._ds = None

    @staticmethod
    def _check_dims_coords(dims, static_coords, model):
        expected_coords = [
            "projection_x_coordinate",
            "projection_y_coordinate",
        ]
        if model == "mo-atmospheric-mogreps-uk":
            expected_coords += ["realization"]

        expected_dims = [
            "forecast_reference_time",
            "forecast_period",
        ] + expected_coords

        pair_dict = {
            "static_coords": (expected_coords, list(static_coords.keys())),
            "dims": (expected_dims, dims),
        }

        for var_type, pair in pair_dict.items():
            expected, passed_in = pair
            # check all of expected were passed in
            if not all(map(lambda var: var in passed_in, expected)):
                raise ValueError(f"Expected to find all of {expected} in {var_type}")

    def _validate_storage_options(self):
        ABFS_KEYS = ("account_name", "credential")
        if "abfs" not in self.data_protocol:
            return

        try:
            options = (
                self.storage_options["abfs"]
                if "::" in self.data_protocol
                else self.storage_options
            )
        except KeyError as err:
            msg = (
                "Expected to find 'abfs' key in storage_options for nested filesystems. "
                f"The associated value should be a dict containing the keys: {ABFS_KEYS}"
            )
            raise KeyError(msg) from err

        # need to check abfs keys are there:
        missing_keys = map(lambda k: k not in options, ABFS_KEYS)
        if any(missing_keys):
            msg = f"When using 'abfs', storage_options should contain the keys: {ABFS_KEYS}"
            raise KeyError(msg)

    @property
    def static_coords(self):
        static_coords = {}
        for name, defn in self._static_coords.items():
            data = defn["data"]
            if isinstance(data, dict):
                data = np.linspace(**data)
            else:
                data = np.array(data)
            static_coords[name] = xr.Variable(
                dims=(name,), data=data, attrs=defn.get("attrs")
            )
        return static_coords

    @property
    def coord_vars(self):
        return dict(self.static_coords, **self.dynamic_coords)

    @property
    def dynamic_coords(self):
        dynamic_coords_data = {
            "forecast_reference_time": pd.date_range(
                start=self.start_cycle, end=self.end_cycle, freq=self.cycle_freq
            ),
            "forecast_period": pd.timedelta_range(
                start=self.start_lead_time,
                end=self.end_lead_time,
                freq=self.lead_time_freq,
            ),
        }
        return {
            name: xr.Variable(dims=(name,), data=data)
            for name, data in dynamic_coords_data.items()
        }

    @property
    def chunks(self):
        static_coords = self.static_coords
        return {name: static_coords[name].shape[0] for name in static_coords.keys()}

    @staticmethod
    def _extract_data_as_dataarray(dataset):
        # coords in all datasets
        REQUIRED_COORD_VARS = [
            "time",
            "forecast_reference_time",
            "realization",
            "projection_y_coordinate",
            "forecast_period",
            "projection_x_coordinate",
        ]

        # coords only in some datasets
        VERTICAL_COORD_VARS = ["height", "depth", "pressure"]

        # variables that are not in coords or dims (including 'bnds' in dims)
        # but in all datasets
        REQUIRED_NON_COORD_DIM_VARS = [
            "lambert_azimuthal_equal_area",
            "projection_y_coordinate_bnds",
            "projection_x_coordinate_bnds",
        ]

        # vars that are not coords/dims but are optional
        TIME_BOUND_VARS = ["time_bnds", "forecast_period_bnds"]

        # also has 'depth_bnds' optionally..
        # need to figure out which variable is the data variable
        NON_DATA_VARS = (
            REQUIRED_COORD_VARS
            + VERTICAL_COORD_VARS
            + REQUIRED_NON_COORD_DIM_VARS
            + TIME_BOUND_VARS
            + ["depth_bnds"]
        )

        data_var = list(set(dataset.variables) - set(NON_DATA_VARS))
        if len(data_var) > 1:
            raise RuntimeError("Expected to find only 1 data variable")
        data_var = data_var[0]
        return dataset[data_var]

    def _get_url(self, diagnostic, cycle_time=None, validity_time=None, lead_time=None):
        """Return the URL of a forecast file."""

        # determine all times
        cycle_time, validity_time, lead_time = calc_cycle_validity_lead_times(
            cycle_time, validity_time, lead_time
        )

        # convert all to strings
        cycle_time = datetime_to_iso_str(cycle_time)
        validity_time = datetime_to_iso_str(validity_time)
        lead_time = timedelta_to_duration_str(lead_time)

        obj_path = (
            f"{self.model}/{cycle_time}/{validity_time}-{lead_time}-{diagnostic}.nc"
        )
        obj_path = f"{self.url_prefix}/{obj_path}"
        return f"{self.data_protocol}://{obj_path}"

    def _read_from_url(self, url, mode="rb"):
        with fsspec.open(url, mode, **self.storage_options) as of:
            data = of.read()
        return data

    def _zstore_loader(self, attrs):
        ref_time = attrs["forecast_reference_time"]
        fcst_period = attrs["forecast_period"]
        diag = attrs["variable_name"]

        ref_time = pd.to_datetime(np.datetime64(ref_time, "ns"))
        fcst_period = pd.to_timedelta(np.timedelta64(fcst_period, "ns"))

        url = self._get_url(diagnostic=diag, cycle_time=ref_time, lead_time=fcst_period)

        try:
            data = self._read_from_url(url)
            data = xr.open_dataset(BytesIO(data))
            data = self._extract_data_as_dataarray(data)
            return data.values
        except FileNotFoundError:
            return None

    def _create_zstore(self):
        return HypotheticZarrStore(
            dims=self.dims,
            coord_vars=self.coord_vars,
            data_vars=self.diagnostics,
            chunks=self.chunks,
            loader_function=self._zstore_loader,
            attrs=None,
            dtypes=None,
        )

    @property
    def ds(self):
        if self._ds is None:
            self._ds = xr.open_zarr(self._zstore, consolidated=True)
        return self._ds

    def to_xarray(self):
        return self.ds
